#define INCL_BASE
#define INCL_LONGLONG
#include <os2.h>
#include <stdlib.h>
#include <stdio.h>
#include <conio.h>

#define USE_NSPR_CODE

#ifdef USE_NSPR_CODE
//#define PR_INTERVAL_MAX 100000UL
//#define PR_INTERVAL_MAX 0x186a0
#define PR_INTERVAL_MAX 0xffff

ULONG _os2_ticksPerSec = -1;
ULONG _os2_bitShift = 0;
ULONG _os2_highMask = 0;

void _PR_MD_INTERVAL_INIT()
{
  ULONG timerFreq;
  APIRET rc;

  timerFreq = 0; /* OS/2 high-resolution timer frequency in Hz */
  rc = DosTmrQueryFreq(&timerFreq);
  if (NO_ERROR == rc)
  {
    while (timerFreq > PR_INTERVAL_MAX)
    {
      timerFreq >>= 1;
      _os2_bitShift++;
      _os2_highMask = (_os2_highMask << 1)+1;
    }
    _os2_ticksPerSec = timerFreq;
  }
}

ULONG _PR_MD_GET_INTERVAL()
{
  QWORD timestamp;
  ULONG top;

  DosTmrQueryTime(&timestamp);

  /* Sadly, nspr requires the interval to range from 1000 ticks per
   * second to only 100000 ticks per second. DosTmrQueryTime is too
   * high resolution...
   */
  top = timestamp.ulHi & _os2_highMask;
  top = top << (32 - _os2_bitShift);
  timestamp.ulLo = timestamp.ulLo >> _os2_bitShift;
  timestamp.ulLo = timestamp.ulLo + top;
  return timestamp.ulLo;
}

void main(int argc, char **argv)
{
  ULONG ulTmrTime1, ulTmrTime2;
  ULONG ulMsTime1, ulMsTime2;
  ULONG ulTmrDelta;
  ULONG ulMsDelta;
  ULONG ulTmrDeltaMs;
  LONG lDiff;
  LONG lMaxDiff;

  _PR_MD_INTERVAL_INIT();
  printf("_PR_MD_INTERVAL_INIT: TmrFreq=%d ticks/sec highMask=%x bitShift=%x\n",
         _os2_ticksPerSec, _os2_highMask, _os2_bitShift);
  printf("Press ESC to exit\n");

  lMaxDiff = 0;
  ulTmrTime2 = _PR_MD_GET_INTERVAL();
  DosQuerySysInfo(QSV_MS_COUNT, QSV_MS_COUNT, (PVOID)&ulMsTime2, sizeof(ULONG));
  for ( ; ; )
  {
    DosSleep(5000);
    DosSetPriority( PRTYS_THREAD, PRTYC_TIMECRITICAL, PRTYD_MAXIMUM, 0);
    DosQuerySysInfo(QSV_MS_COUNT, QSV_MS_COUNT, (PVOID)&ulMsTime1, sizeof(ULONG));
    ulTmrTime1 = _PR_MD_GET_INTERVAL();
    DosSetPriority(PRTYS_THREAD, PRTYC_REGULAR, 0, 0);
    ulTmrDelta = ulTmrTime1 - ulTmrTime2;
    ulMsDelta = ulMsTime1 - ulMsTime2;
    ulTmrDeltaMs = ulTmrDelta*1000/_os2_ticksPerSec;
    //ulTmrDeltaMs = ulTmrDelta/(ulTmrFreq/1000);
    lDiff = ulTmrDeltaMs-ulMsDelta;
    if (abs(lDiff) > abs(lMaxDiff)) lMaxDiff = lDiff;
    if (abs(lDiff) > 4) DosBeep(660, 250);
    printf("Tmr=%x Dlt=%x (%4dms) Ms=%x Dlt=%x (%4dms) Dif=%2d Mx=%2d\n",
      ulTmrTime1, ulTmrDelta, ulTmrDeltaMs,
      ulMsTime1, ulMsDelta, ulMsDelta, lDiff, lMaxDiff);
    ulTmrTime2 = ulTmrTime1;
    ulMsTime2 = ulMsTime1;
    if (kbhit() && (getch() == 0x1b)) break;
  }
}
#else
void main(int argc, char **argv)
{
  APIRET rc;
  ULONG ulTmrFreq;
  ULONGLONG qwTmrTime1, qwTmrTime2;
  ULONG ulMsTime1, ulMsTime2;
  ULONG ulTmrDelta;
  ULONG ulMsDelta;
  ULONG ulTmrDeltaMs;
  LONG lDiff;
  LONG lMaxDiff;

  rc = DosTmrQueryFreq(&ulTmrFreq);
  printf("DosTmrQueryFreq: TmrFreq=%d ticks/sec\n", ulTmrFreq);
  printf("Press ESC to exit\n");

  lMaxDiff = 0;
  rc = DosTmrQueryTime((QWORD*)&qwTmrTime2);
  DosQuerySysInfo(QSV_MS_COUNT, QSV_MS_COUNT, (PVOID)&ulMsTime2, sizeof(ULONG));
  for ( ; ; )
  {
    DosSleep(3000);
    DosSetPriority( PRTYS_THREAD, PRTYC_TIMECRITICAL, PRTYD_MAXIMUM, 0);
    DosQuerySysInfo(QSV_MS_COUNT, QSV_MS_COUNT, (PVOID)&ulMsTime1, sizeof(ULONG));
    rc = DosTmrQueryTime((QWORD*)&qwTmrTime1);
    DosSetPriority(PRTYS_THREAD, PRTYC_REGULAR, 0, 0);
    ulTmrDelta = qwTmrTime1 - qwTmrTime2;
    ulMsDelta = ulMsTime1 - ulMsTime2;
    ulTmrDeltaMs = ulTmrDelta*1000/ulTmrFreq;
    //ulTmrDeltaMs = ulTmrDelta/(ulTmrFreq/1000);
    lDiff = ulTmrDeltaMs-ulMsDelta;
    if (abs(lDiff) > abs(lMaxDiff)) lMaxDiff = lDiff;
    if (abs(lDiff) > 4) DosBeep(660, 250);
    printf("Tmr=%llx Dlt=%x (%4dms) Ms=%x Dlt=%x (%4dms) Dif=%2d Mx=%2d\n",
      qwTmrTime1, ulTmrDelta, ulTmrDeltaMs,
      ulMsTime1, ulMsDelta, ulMsDelta, lDiff, lMaxDiff);
    qwTmrTime2 = qwTmrTime1;
    ulMsTime2 = ulMsTime1;
    if (kbhit() && (getch() == 0x1b)) break;
  }
}
#endif
